// トランジスタ技術2020年8月号
// ArduinoとPythonで動かしながら学ぶモータ制御入門
// リスト１（Arduinoスケッチ）
// Y.Minami
//////////////////////////////////////////////
#define PIN_ENCA 2 // Encoder
#define PIN_ENCB 3 // Encoder
#define PIN_IN1  7 // Motor Rotation
#define PIN_IN2  8 // Motor Rotation
#define PIN_VREF 9 // Motor PWM
#define PIN_STB 10 // Motor Standby

const float REF = 30.0;  // 目標値 [deg]
const float Ts = 0.01;   // サンプリング時間 [s]
const float VREF = 4.0;  // 電圧の最大値 [V]
const int PWM_MAX = 255; // PWM 100%
const int PWM_MIN = 0;   // PWM 0%

volatile unsigned long time; // [msec]
volatile uint8_t prev = 0;
volatile int value = 0; // エンコーダのパルス数 [plus]
volatile float ang = 0; // モータ回転角 [deg]

volatile float u = 0.0;  // 制御入力 [V]
volatile float t = 0.0;  // 時間 [s]

void setup() {
  pinMode(PIN_IN1,OUTPUT); 
  pinMode(PIN_IN2,OUTPUT); 
  pinMode(PIN_ENCA, INPUT); 
  pinMode(PIN_ENCB, INPUT);
  digitalWrite(PIN_ENCA, HIGH);
  digitalWrite(PIN_ENCB, HIGH);
  attachInterrupt(0, updateEncoder, CHANGE);
  attachInterrupt(1, updateEncoder, CHANGE);
  Serial.begin(115200);
}

// エンコーダのカウントアップ・カウントダウン
void updateEncoder()
{
  uint8_t a = digitalRead(PIN_ENCA);
  uint8_t b = digitalRead(PIN_ENCB);
  uint8_t ab = (a << 1) | b;
  uint8_t encoded  = (prev << 2) | ab;

  if(encoded == 0b1101 || encoded == 0b0100 || encoded == 0b0010 || encoded == 0b1011){
    value ++;
  } else if(encoded == 0b1110 || encoded == 0b0111 || encoded == 0b0001 || encoded == 0b1000) {
    value --;
  }
  prev = ab;
}

void loop() {

  time = millis();

  // Pythonからのデータを受信
  if ( Serial.available() > 0 ) { 
    String buf = Serial.readStringUntil('e');
    if (buf.length() > 0) {
      u = atof(strtok(buf.c_str(), "," ));
    } 
  }

  // 電圧をPWM(-255〜255)に変換
  u = u * PWM_MAX/VREF;

  // モータの回転方向の切り替えとPWM信号の出力
  int mag = constrain(abs(u), PWM_MIN, PWM_MAX);
  digitalWrite(PIN_STB, HIGH);
  if(u >= 0){
    analogWrite(PIN_VREF, mag); 
    digitalWrite(PIN_IN1, HIGH);
    digitalWrite(PIN_IN2, LOW);
  }
  else{
    analogWrite(PIN_VREF, mag); 
    digitalWrite(PIN_IN1, LOW);
    digitalWrite(PIN_IN2, HIGH);
  }

  // 回転角度の計算
  // モータ側：1回転で12パルス，ギア比が1:120なので，出力1回転は 1440[plus]
  // したがって，1 パルスあたり　360[deg] / 1440[plus] = 0.25 [deg/plus]  
  ang = float(value) * 0.25;

  while(1){ // Ts[s]ごとに繰り返す
    if( (millis() - time) >= Ts * 1000 )
    break;
  }
  t = t + Ts;

  // Pythonへデータを送信
  String str = String(t) + " " + String(ang, 3) + " "+ String(u*100/PWM_MAX, 3) + " ";
  Serial.println(str); 

}
